import asyncio
import logging
from datetime import datetime, timedelta
from functools import wraps
from typing import TYPE_CHECKING, Optional

from cashews._typing import TTL, AsyncCallable_T, CallableCacheCondition, Decorator, KeyOrTemplate, Tags
from cashews.formatter import register_template
from cashews.key import get_cache_key, get_cache_key_template
from cashews.ttl import ttl_to_seconds

from .defaults import _empty, context_cache_detect

if TYPE_CHECKING:  # pragma: no cover
    from cashews.backends import Cache

__all__ = ("early",)


logger = logging.getLogger(__name__)
_LOCK_SUFFIX = ":lock"


def early(
    backend: "Cache",
    ttl: TTL,
    key: Optional[KeyOrTemplate] = None,
    early_ttl: Optional[TTL] = None,
    condition: CallableCacheCondition = lambda *args, **kwargs: True,
    prefix: str = "early",
    tags: Tags = (),
) -> Decorator:
    """
    Cache strategy that try to solve Cache stampede problem (https://en.wikipedia.org/wiki/Cache_stampede),
    With hot cache recalculate a result in background near expiration time
    Warning Not good at cold cache

    :param backend: cache backend
    :param ttl: duration in seconds to store a result
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param early_ttl: duration in seconds to expire results
    :param condition: callable object that determines whether the result will be saved or not
    :param prefix: custom prefix for key, default 'early'
    :param tags: aliases for keys that used for cache (used for invalidation)
    """

    ttl = ttl_to_seconds(ttl)
    early_ttl = ttl_to_seconds(early_ttl)

    def _decor(func: AsyncCallable_T) -> AsyncCallable_T:
        _key_template = get_cache_key_template(func, key=key, prefix=prefix + ":v2")
        register_template(func, _key_template)
        for tag in tags:
            backend.register_tag(tag, _key_template)

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _ttl = ttl_to_seconds(ttl, *args, **kwargs, with_callable=True)
            if early_ttl is None:
                _early_ttl = _ttl * 0.33
            else:
                _early_ttl = ttl_to_seconds(early_ttl, *args, **kwargs, with_callable=True)
            _cache_key = get_cache_key(func, _key_template, args, kwargs)
            _tags = [get_cache_key(func, tag, args, kwargs) for tag in tags]

            args_to_call = [backend, func, args, kwargs, _cache_key, _ttl, _early_ttl, condition, _tags]
            cached = await backend.get(_cache_key, default=_empty)
            if cached is _empty:
                return await _get_result_for_early(*args_to_call)

            context_cache_detect._set(
                _cache_key,
                ttl=_ttl,
                early_ttl=_early_ttl,
                name="early",
                template=_key_template,
            )
            early_expire_at, result = cached
            if early_expire_at >= datetime.utcnow():
                return result
            lock_key = _cache_key + _LOCK_SUFFIX
            if not await backend.set(lock_key, "1", expire=_early_ttl, exist=False):
                return result
            logger.info(
                "Recalculate cache for %s (exp_at %s)",
                _cache_key,
                early_expire_at,
            )
            asyncio.create_task(_get_result_for_early(*args_to_call, unlock=True))
            return result

        return _wrap

    return _decor


async def _get_result_for_early(
    backend: "Cache", func, args, kwargs, key, ttl: int, early_ttl: int, condition, tags, unlock=False
):
    try:
        result = await func(*args, **kwargs)
        if condition(result, args, kwargs, key):
            early_expire_at = datetime.utcnow() + timedelta(seconds=early_ttl)
            await backend.set(key, [early_expire_at, result], expire=ttl, tags=tags)
        return result
    finally:
        if unlock:
            asyncio.create_task(backend.delete(key + _LOCK_SUFFIX))



from contextlib import contextmanager
from typing import TYPE_CHECKING

from cashews._typing import AsyncCallable_T
from cashews.commands import Command

from .wrapper import Wrapper

if TYPE_CHECKING:  # pragma: no cover
    from cashews.backends.interface import Backend


async def _is_disable_middleware(call: AsyncCallable_T, cmd: Command, backend: "Backend", *args, **kwargs):
    if backend.is_disable(cmd):
        if cmd in (Command.GET, Command.GET_MANY):
            return kwargs.get("default", None)
        return None
    return await call(*args, **kwargs)


class ControlWrapper(Wrapper):
    def __init__(self, name: str = ""):
        super().__init__(name)
        self.add_middleware(_is_disable_middleware)

    def disable(self, *cmds: Command, prefix: str = "") -> None:
        return self._get_backend(prefix).disable(*cmds)

    def enable(self, *cmds: Command, prefix: str = "") -> None:
        return self._get_backend(prefix).enable(*cmds)

    @contextmanager
    def disabling(self, *cmds: Command, prefix: str = ""):
        self.disable(*cmds, prefix=prefix)
        try:
            yield
        finally:
            self.enable(*cmds, prefix=prefix)

    def is_disable(self, *cmds: Command, prefix: str = ""):
        return self._get_backend(prefix).is_disable(*cmds)

    def is_enable(self, *cmds: Command, prefix: str = ""):
        return not self.is_disable(*cmds, prefix=prefix)

    @property
    def is_full_disable(self):
        return all([backend.is_full_disable for backend, _ in self._backends.values()])



import asyncio
import uuid
from abc import ABCMeta, abstractmethod
from contextlib import asynccontextmanager
from contextvars import ContextVar
from typing import Any, AsyncIterator, Iterable, List, Mapping, Optional, Set, Tuple

from cashews._typing import Callback, Key, Value
from cashews.commands import ALL, Command
from cashews.exceptions import CacheBackendInteractionError, LockedError

NOT_EXIST = -2
UNLIMITED = -1


class _BackendInterface(metaclass=ABCMeta):
    @property
    @abstractmethod
    def is_init(self) -> bool:
        ...

    @abstractmethod
    async def init(self):
        ...

    @abstractmethod
    async def close(self):
        ...

    @abstractmethod
    async def set(
        self,
        key: Key,
        value: Value,
        expire: Optional[float] = None,
        exist: Optional[bool] = None,
    ) -> bool:
        ...

    @abstractmethod
    async def set_many(self, pairs: Mapping[Key, Value], expire: Optional[float] = None):
        ...

    @abstractmethod
    async def set_raw(self, key: Key, value: Value, **kwargs: Any):
        ...

    @abstractmethod
    async def get(self, key: Key, default: Optional[Value] = None) -> Value:
        ...

    @abstractmethod
    async def get_raw(self, key: Key) -> Value:
        ...

    @abstractmethod
    async def get_many(self, *keys: Key, default: Optional[Value] = None) -> Tuple[Optional[Value], ...]:
        ...

    @abstractmethod
    async def get_match(self, pattern: str, batch_size: int = 100) -> AsyncIterator[Tuple[Key, Value]]:
        ...

    @abstractmethod
    async def exists(self, key: Key) -> bool:
        ...

    @abstractmethod
    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[Key]:
        ...

    @abstractmethod
    async def incr(self, key: Key, value: int = 1, expire: Optional[float] = None) -> int:
        ...

    @abstractmethod
    async def delete(self, key: Key) -> bool:
        ...

    @abstractmethod
    async def delete_many(self, *keys: Key):
        ...

    @abstractmethod
    async def delete_match(self, pattern: str):
        ...

    @abstractmethod
    async def expire(self, key: Key, timeout: float):
        ...

    @abstractmethod
    async def get_expire(self, key: Key) -> int:
        ...

    @abstractmethod
    async def get_bits(self, key: Key, *indexes: int, size: int = 1) -> Tuple[int, ...]:
        ...

    @abstractmethod
    async def incr_bits(self, key: Key, *indexes: int, size: int = 1, by: int = 1) -> Tuple[int, ...]:
        ...

    @abstractmethod
    async def slice_incr(self, key: Key, start: int, end: int, maxvalue: int, expire: Optional[float] = None) -> int:
        ...

    @abstractmethod
    async def set_add(self, key: Key, *values: str, expire: Optional[float] = None):
        ...

    @abstractmethod
    async def set_remove(self, key: Key, *values: str):
        ...

    @abstractmethod
    async def set_pop(self, key: Key, count: int = 100) -> Iterable[str]:
        ...

    @abstractmethod
    async def get_size(self, key: Key) -> int:
        """
        Return size in bites that allocated by a value for given key
        """
        ...

    @abstractmethod
    async def ping(self, message: Optional[bytes] = None) -> bytes:
        ...

    @abstractmethod
    async def clear(self):
        ...

    async def set_lock(self, key: Key, value: Value, expire: float) -> bool:
        return await self.set(key, value, expire=expire, exist=False)

    @abstractmethod
    async def is_locked(
        self,
        key: Key,
        wait: Optional[float] = None,
        step: float = 0.1,
    ) -> bool:
        ...

    @abstractmethod
    async def unlock(self, key: Key, value: Value) -> bool:
        ...

    @asynccontextmanager
    async def lock(self, key: Key, expire: float, wait=True):
        identifier = str(uuid.uuid4())
        while True:
            lock = await self.set_lock(key, identifier, expire=expire)
            if not lock:
                # we need to check the connection by ping because
                # if redis unavailable and a backend have flag `safe`
                # we will have a brake lock
                try:
                    if await self.ping(b"LOCK") is None:
                        yield
                        return
                except CacheBackendInteractionError:
                    yield
                    return

                if wait:
                    await asyncio.sleep(0)
                    continue
                raise LockedError(f"Key {key} is already locked")
            try:
                yield
            finally:
                await self.unlock(key, identifier)
            return


class ControlMixin:
    def __init__(self, *args, **kwargs) -> None:
        self.__disable: ContextVar[Set[Command]] = ContextVar(str(id(self)), default=set())
        super().__init__(*args, **kwargs)

    @property
    def _disable(self) -> Set[Command]:
        return self.__disable.get()

    def _set_disable(self, value: Set[Command]) -> None:
        self.__disable.set(value)

    def is_disable(self, *cmds: Command) -> bool:
        _disable = self._disable
        if not cmds and _disable:
            return True
        for cmd in cmds:
            if cmd in _disable:
                return True
        return False

    def is_enable(self, *cmds: Command) -> bool:
        return not self.is_disable(*cmds)

    @property
    def is_full_disable(self):
        return self._disable == ALL

    def disable(self, *cmds: Command) -> None:
        if not cmds:
            _disable = ALL.copy()
        else:
            _disable = self._disable.copy()
            _disable.update(cmds)
        self._set_disable(_disable)

    def enable(self, *cmds: Command) -> None:
        if not cmds:
            _disable = set()
        else:
            _disable = self._disable.copy()
            _disable -= set(cmds)
        self._set_disable(_disable)


class Backend(ControlMixin, _BackendInterface, metaclass=ABCMeta):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._on_remove_callbacks: List[Callback] = []

    def on_remove_callback(self, callback: Callback):
        self._on_remove_callbacks.append(callback)

    async def _call_on_remove_callbacks(self, *keys: Key):
        for callback in self._on_remove_callbacks:
            await callback(keys, backend=self)



import asyncio
import re
from datetime import datetime
from typing import Any, AsyncIterator, Dict, Iterable, List, Mapping, Optional, Tuple

from diskcache import Cache, FanoutCache

from cashews._typing import Key, Value
from cashews.serialize import SerializerMixin
from cashews.utils import Bitarray

from .interface import NOT_EXIST, UNLIMITED, Backend


class _DiskCache(Backend):
    def __init__(self, *args, directory=None, shards=8, **kwargs: Any) -> None:
        self.__is_init = False
        self._set_locks: Dict[str, asyncio.Lock] = {}
        self._sharded = shards > 1
        if not self._sharded:
            self._cache = Cache(directory=directory, **kwargs)
        else:
            self._cache = FanoutCache(directory=directory, shards=shards, **kwargs)
        super().__init__(**kwargs)

    async def init(self):
        self.__is_init = True

    async def _run_in_executor(self, call, *args):
        return await asyncio.get_running_loop().run_in_executor(None, call, *args)

    @property
    def is_init(self) -> bool:
        return self.__is_init

    async def close(self):
        self._cache.close()
        self.__is_init = False

    async def set(
        self,
        key: Key,
        value: Value,
        expire: Optional[float] = None,
        exist: Optional[bool] = None,
    ) -> bool:
        future = self._run_in_executor(self._set, key, value, expire, exist)
        if exist is not None:
            # we should have async lock until value real set
            lock = self._set_locks.setdefault(key, asyncio.Lock())
            async with lock:
                try:
                    return await future
                finally:
                    self._set_locks.pop(key, None)
        return await future

    def _set(self, key: Key, value: Value, expire=None, exist=None):
        if exist is not None:
            if not self._exists(key) is exist:
                return False
        if expire is None:
            expire = self._get_expire(key)
            expire = expire if expire not in [UNLIMITED, NOT_EXIST] else None
        return self._cache.set(key, value, expire)

    async def set_raw(self, key: Key, value: Any, **kwargs: Any):
        return self._cache.set(key, value, **kwargs)

    async def get(self, key: Key, default: Optional[Value] = None) -> Value:
        return await self._run_in_executor(self._cache.get, key, default)

    async def get_raw(self, key: Key) -> Value:
        return self._cache.get(key)

    async def get_many(self, *keys: Key, default: Optional[Value] = None) -> Tuple[Value]:
        return await self._run_in_executor(self._get_many, keys, default)

    def _get_many(self, keys: List[Key], default: Optional[Value] = None):
        values = []
        for key in keys:
            val = self._cache.get(key, default=default)
            if isinstance(val, Bitarray):
                val = None
            values.append(val)
        return values

    async def set_many(self, pairs: Mapping[Key, Value], expire: Optional[float] = None):
        return await self._run_in_executor(self._set_many, pairs, expire)

    def _set_many(self, pairs: Mapping[Key, Value], expire: Optional[float] = None):
        for key, value in pairs.items():
            self._set(key, value, expire=expire)

    async def exists(self, key: Key) -> bool:
        return await self._run_in_executor(self._exists, key)

    def _exists(self, key: Key) -> bool:
        return key in self._cache

    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[Value]:  # type: ignore
        if not self._sharded:
            for key in await self._run_in_executor(self._scan, pattern):
                yield key

    def _scan(self, pattern: str) -> Value:
        pattern = pattern.replace("*", ".*")
        regexp = re.compile(pattern)
        for key in self._cache.iterkeys():
            if regexp.fullmatch(key):
                yield key

    async def get_bits(self, key: Key, *indexes: int, size: int = 1) -> Tuple[int, ...]:
        array = await self.get(key, default=Bitarray("0"))
        result = []
        for index in indexes:
            result.append(array.get(index, size))
        return tuple(result)

    async def incr_bits(self, key: str, *indexes: int, size: int = 1, by: int = 1) -> Tuple[int, ...]:
        array = await self.get(key, default=Bitarray("0"))
        result = []
        for index in indexes:
            array.incr(index, size, by)
            result.append(array.get(index, size))
        await self.set(key, array)
        return tuple(result)

    async def incr(self, key: Key, value: int = 1, expire: Optional[float] = None) -> int:
        return await self._run_in_executor(self._incr, key, value, expire)

    def _incr(self, key: Key, value: int = 1, expire: Optional[float] = None) -> int:
        res = self._cache.incr(key, delta=value, retry=True)
        if res == 1 and expire:
            self._cache.touch(key, expire)
        return res

    async def delete(self, key: Key) -> bool:
        try:
            return await self._run_in_executor(self._cache.delete, key)
        finally:
            await self._call_on_remove_callbacks(key)

    async def delete_many(self, *keys: Key):
        try:
            await self._run_in_executor(self._delete_many, keys)
        finally:
            await self._call_on_remove_callbacks(*keys)

    def _delete_many(self, keys: List[Key]):
        for key in keys:
            self._cache.delete(key)

    async def delete_match(self, pattern: str):
        keys = []
        async for key in self.scan(pattern):
            keys.append(key)
        try:
            return await self._run_in_executor(self._delete_match, pattern)
        finally:
            await self._call_on_remove_callbacks(*keys)

    def _delete_match(self, pattern: str):
        for key in self._scan(pattern):
            self._cache.delete(key)

    async def get_match(self, pattern: str, batch_size: int = None) -> AsyncIterator[Tuple[Key, Value]]:
        if self._sharded:
            return
        for key in await self._run_in_executor(self._scan, pattern):
            value = await self.get(key)
            if isinstance(value, Bitarray):
                continue
            yield key, value

    async def expire(self, key: Key, timeout: float) -> int:
        return await self._run_in_executor(self._cache.touch, key, timeout)

    async def get_expire(self, key: Key) -> int:
        return await self._run_in_executor(self._get_expire, key)

    def _get_expire(self, key: Key) -> int:
        value, expire = self._cache.get(key, expire_time=True)
        if value is None:
            return NOT_EXIST
        if expire is None:
            return UNLIMITED
        return round((datetime.utcfromtimestamp(expire) - datetime.utcnow()).total_seconds())

    async def get_size(self, key: Key) -> int:
        return -1

    async def ping(self, message: Optional[bytes] = None) -> bytes:
        return b"PONG" if message in (None, b"PING") else message

    async def clear(self):
        await self._run_in_executor(self._cache.clear)

    async def is_locked(
        self,
        key: Key,
        wait: Optional[float] = None,
        step: float = 0.1,
    ) -> bool:
        if wait is None:
            return await self.exists(key)
        while wait > 0:
            if not await self.exists(key):
                return False
            wait -= step
            await asyncio.sleep(step)
        return await self.exists(key)

    async def unlock(self, key: Key, value: Value) -> bool:
        return await self._run_in_executor(self._unlock, key, value)

    def _unlock(self, key: Key, value: Value) -> bool:
        if self._cache.get(key) == value:
            self._cache.delete(key)
            return True
        return False

    async def slice_incr(self, key: Key, start: int, end: int, maxvalue: int, expire: Optional[float] = None) -> int:
        val_set = await self.get(key)
        count = 0
        new_val = []
        if val_set:
            for val in val_set:
                if start <= val <= end:
                    count += 1
                    new_val.append(val)

        if count < maxvalue:
            count += 1
            new_val.append(end)
        await self.set(key, new_val, expire=expire)
        return count

    async def set_add(self, key: Key, *values: str, expire: Optional[float] = None):
        val = await self.get(key, default=set())
        val.update(values)
        await self.set(key, val, expire=expire)

    async def set_remove(self, key: Key, *values: str):
        val = await self.get(key, default=set())
        val.difference_update(values)
        await self.set(key, val)

    async def set_pop(self, key: Key, count: int = 100) -> Iterable[str]:
        values = await self.get(key, default=set())
        _values = []
        for _ in range(count):
            if not values:
                break
            _values.append(values.pop())

        await self.set(key, values)
        return _values


class DiskCache(SerializerMixin, _DiskCache):
    pass



from cashews.backends.interface import _BackendInterface
from cashews.decorators import context_cache_detect

from .backend_settings import register_backend  # noqa
from .decorators import DecoratorsWrapper
from .disable_control import ControlWrapper
from .tags import CommandsTagsWrapper
from .transaction import TransactionMode, TransactionWrapper

__all__ = ["Cache", "register_backend"]


class Cache(
    TransactionWrapper,
    ControlWrapper,
    CommandsTagsWrapper,
    DecoratorsWrapper,
    _BackendInterface,
):
    detect = context_cache_detect



from functools import partial
from typing import Dict, List, Tuple

from cashews import validation
from cashews._typing import Key, Middleware
from cashews.backends.interface import Backend
from cashews.commands import Command
from cashews.exceptions import NotConfiguredError

from .auto_init import create_auto_init
from .backend_settings import settings_url_parse


class Wrapper:
    default_prefix = ""

    def __init__(self, name: str = ""):
        self._backends: Dict[str, Tuple[Backend, Tuple[Middleware, ...]]] = {}
        self._default_middlewares: List[Middleware, ...] = [
            create_auto_init(),
            validation._invalidate_middleware,
        ]
        self.name = name
        super().__init__()

    def add_middleware(self, middleware: Middleware):
        self._default_middlewares.append(middleware)

    def _get_backend_and_config(self, key: Key) -> Tuple[Backend, Tuple[Middleware, ...]]:
        for prefix in sorted(self._backends.keys(), reverse=True):
            if key.startswith(prefix):
                return self._backends[prefix]
        if self.default_prefix not in self._backends:
            raise NotConfiguredError("run `cache.setup(...)` before using cache")

    def _get_backend(self, key: Key) -> Backend:
        backend, _ = self._get_backend_and_config(key)
        return backend

    def _with_middlewares(self, cmd: Command, key: Key):
        backend, middlewares = self._get_backend_and_config(key)
        return self._with_middlewares_for_backend(cmd, backend, middlewares)

    def _with_middlewares_for_backend(self, cmd: Command, backend, middlewares):
        call = getattr(backend, cmd.value)
        for middleware in middlewares:
            call = partial(middleware, call, cmd, backend)
        return call

    def setup(self, settings_url: str, middlewares: Tuple = (), prefix: str = default_prefix, **kwargs) -> Backend:
        backend_class, params = settings_url_parse(settings_url)
        params.update(kwargs)

        if "disable" in params:
            disable = params.pop("disable")
        else:
            disable = not params.pop("enable", True)

        backend = backend_class(**params)
        if disable:
            backend.disable()
        self._add_backend(backend, middlewares, prefix)
        return backend

    def _add_backend(self, backend: Backend, middlewares=(), prefix: str = default_prefix):
        self._backends[prefix] = (
            backend,
            tuple(self._default_middlewares) + middlewares,
        )

    async def init(self, *args, **kwargs):
        if args or kwargs:
            self.setup(*args, **kwargs)
        for backend, _ in self._backends.values():
            await backend.init()

    @property
    def is_init(self) -> bool:
        for backend, _ in self._backends.values():
            if not backend.is_init:
                return False
        return True

    async def close(self):
        for backend, _ in self._backends.values():
            await backend.close()



from cashews.picklers import DEFAULT_PICKLE
from cashews.serialize import SerializerMixin

from .backend import _Redis

__all__ = ["Redis"]


class Redis(SerializerMixin, _Redis):
    pickle_type = DEFAULT_PICKLE



import hashlib
import hmac
import warnings
from typing import TYPE_CHECKING, Any, Mapping, Optional, Tuple, Type, Union

from ._typing import Key, Value
from .exceptions import SignIsMissingError, UnSecureDataError
from .picklers import DEFAULT_PICKLE, NULL_PICKLE, Pickler, get_pickler

if TYPE_CHECKING:
    from .backends.interface import Backend

_empty = object()


def _seal(digestmod):
    def sign(key: bytes, value: bytes) -> bytes:
        return hmac.new(key, value, digestmod).hexdigest().encode()

    return sign


def simple_sign(key: bytes, value: bytes) -> bytes:
    s = sum(key) + sum(value)
    return f"{s:x}".encode()


class SerializerMixin:
    pickle_type = NULL_PICKLE

    def __init__(
        self,
        *args,
        hash_key: Union[str, bytes, None] = _empty,
        secret: Union[str, bytes, None] = None,
        digestmod: Union[str, bytes, None] = b"md5",
        check_repr: bool = True,
        pickle_type: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        if hash_key is not _empty:
            warnings.warn(
                "`hash_key` property was renamed to `secret` and will be removed in next release",
                DeprecationWarning,
            )
            secret = hash_key

        self._serializer = Serializer(check_repr=check_repr)
        if secret:
            self._serializer.set_signer(HashSigner(secret, digestmod))

        self._serializer.set_pickler(self._get_pickler(pickle_type, secret))

    @classmethod
    def _get_pickler(cls, pickle_type: Optional[str], hash_key: bool) -> Pickler:
        pickle_type = pickle_type or cls.pickle_type
        if pickle_type is NULL_PICKLE and hash_key:
            pickle_type = DEFAULT_PICKLE
        return get_pickler(pickle_type)

    async def get(self, key: Key, default: Optional[Value] = None):
        raw_value = await super().get(key, default=default)
        return await self._serializer.decode(self, key, raw_value, default=default)

    async def get_many(self, *keys: Key, default: Optional[Value] = None) -> Value:
        encoded_values = await super().get_many(*keys, default=default)
        values = []
        for key, value in zip(keys, encoded_values):
            deserialized_value = await self._serializer.decode(self, key, value, default=default)
            values.append(deserialized_value)
        return tuple(values)

    async def set(
        self,
        key: Key,
        value: Value,
        expire: Optional[float] = None,
        exist: Optional[bool] = None,
    ):
        value = await self._serializer.encode(self, key, value, expire=expire)
        return await super().set(key, value, expire=expire, exist=exist)

    async def set_many(self, pairs: Mapping[Key, Value], expire: Optional[float] = None):
        transformed_pairs = {}
        for key, value in pairs.items():
            transformed_pairs[key] = await self._serializer.encode(self, key, value, expire)
        return await super().set_many(transformed_pairs, expire=expire)

    def set_raw(self, *args: Any, **kwargs: Any):
        return super().set(*args, **kwargs)

    def get_raw(self, *args: Any, **kwargs: Any):
        return super().get(*args, **kwargs)


def _to_bytes(value: Union[str, bytes, None]) -> Optional[bytes]:
    if value is None:
        return None
    if isinstance(value, str):
        value = value.encode()
    return value


class HashSigner:
    _digestmods = {
        b"sha1": _seal(hashlib.sha1),
        b"md5": _seal(hashlib.md5),
        b"sha256": _seal(hashlib.sha256),
        b"sum": simple_sign,
    }

    def __init__(self, secret: Union[str, bytes], digestmod: Union[str, bytes] = b"md5"):
        self._secret = _to_bytes(secret)
        self._digestmod = _to_bytes(digestmod)

    def sign(self, key: Key, value: bytes) -> bytes:
        sign = self._gen_sign(key, value, self._digestmod)
        return self._digestmod + b":" + sign + b"_" + value

    def check_sign(self, key: Key, value: bytes) -> bytes:
        try:
            sign, value = value.split(b"_", 1)
        except ValueError as exc:
            raise SignIsMissingError(f"key: {key}") from exc

        sign, digestmod = self._get_sign_and_digestmod(sign)
        expected_sign = self._gen_sign(key, value, digestmod)
        if expected_sign != sign:
            raise UnSecureDataError(f"{expected_sign!r} != {sign!r}")
        return value

    def _gen_sign(self, key: Key, value: bytes, digestmod: bytes) -> bytes:
        value = key.encode() + value
        return self._digestmods[digestmod](self._secret, value)

    def _get_sign_and_digestmod(self, sign: bytes) -> Tuple[bytes, bytes]:
        digestmod = self._digestmod
        if b":" in sign:
            digestmod, sign = sign.split(b":")
        if digestmod not in self._digestmods:
            raise UnSecureDataError()
        return sign, digestmod


class NullSigner:
    @staticmethod
    def sign(key: Key, value: bytes) -> bytes:
        return value

    @staticmethod
    def check_sign(key: Key, value: bytes) -> bytes:
        return value


class Serializer:
    _type_mapping = {}

    def __init__(self, check_repr=False):
        self._check_repr = check_repr
        self._pickler = get_pickler(NULL_PICKLE)
        self._signer = NullSigner()

    def set_signer(self, signer):
        self._signer = signer

    def set_pickler(self, pickler):
        self._pickler = pickler

    @classmethod
    def register_type(cls, klass: Type, encoder, decoder):
        cls._type_mapping[bytes(klass.__name__, "utf8")] = (encoder, decoder)

    async def encode(self, backend: "Backend", key: Key, value: Value, expire: int) -> bytes:  # on SET
        if isinstance(value, int) and not isinstance(value, bool):
            return value
        _value = await self._custom_encode(backend, key, value, expire)
        if _value is not None:
            return self._signer.sign(key, _value)
        return self._signer.sign(key, self._pickler.dumps(value))

    async def _custom_encode(self, backend, key: Key, value: Value, expire: int) -> Optional[bytes]:
        value_type = bytes(type(value).__name__, "utf8")
        if value_type not in self._type_mapping:
            return
        encoder, _ = self._type_mapping[value_type]
        encoded_value = await encoder(value, backend, key, expire)
        return value_type + b":" + encoded_value

    async def decode(self, backend: "Backend", key: Key, value: bytes, default: Value) -> Value:  # on GET
        if value is default:
            return default
        if not isinstance(value, bytes):
            return value
        if value.isdigit():
            return int(value)
        try:
            value = self._signer.check_sign(key, value)
        except SignIsMissingError:
            return default

        try:
            value = self._decode(value)
        except self._pickler.UnpicklingError:
            pass
        except AttributeError:
            return default
        if isinstance(value, bytes):
            return await self._custom_decode(backend, key, value, default)
        return value

    def _decode(self, value: bytes) -> Value:
        value = self._pickler.loads(value)
        if self._check_repr:
            repr(value)
        return value

    async def _custom_decode(self, backend: "Backend", key: Key, value: bytes, default: Value) -> Value:
        try:
            value_type, value = value.split(b":", 1)
        except ValueError:
            return default
        if value_type not in self._type_mapping:
            return default
        _, decoder = self._type_mapping[value_type]
        decode_value = await decoder(value, backend, key)
        return decode_value


register_type = Serializer.register_type


async def bytes_encoder(value: bytes, *args, **kwargs):
    return value


async def bytes_decoder(value: bytes, *args, **kwargs):
    return value


register_type(bytes, bytes_encoder, bytes_decoder)



import asyncio
import re
import time
from collections import OrderedDict
from contextlib import suppress
from typing import Any, AsyncIterator, Iterable, Mapping, Optional, Tuple

from cashews._typing import Key, Value
from cashews.serialize import SerializerMixin
from cashews.utils import Bitarray, get_obj_size

from .interface import NOT_EXIST, UNLIMITED, Backend

__all__ = ["Memory"]

_missed = object()


class _Memory(Backend):
    """
    Inmemory backend lru with ttl
    """

    __slots__ = ["store", "_check_interval", "size", "__is_init", "__remove_expired_stop", "__remove_expired_task"]

    def __init__(self, size: int = 1000, check_interval: float = 1, **kwargs):
        self.store: OrderedDict = OrderedDict()
        self._check_interval = check_interval
        self.size = size
        self.__is_init = False
        self.__remove_expired_stop = asyncio.Event()
        self.__remove_expired_task = None
        super().__init__(**kwargs)

    async def init(self):
        self.__is_init = True
        if self._check_interval:
            self.__remove_expired_stop = asyncio.Event()
            self.__remove_expired_task = asyncio.create_task(self._remove_expired())

    @property
    def is_init(self) -> bool:
        return self.__is_init

    async def _remove_expired(self):
        while not self.__remove_expired_stop.is_set():
            for key in dict(self.store):
                await self.get(key)
            with suppress(asyncio.TimeoutError, TimeoutError):
                await asyncio.wait_for(self.__remove_expired_stop.wait(), self._check_interval)

    async def clear(self):
        self.store = OrderedDict()

    async def set(
        self,
        key: Key,
        value: Value,
        expire: Optional[float] = None,
        exist: Optional[bool] = None,
    ) -> bool:
        if exist is not None:
            if not (key in self.store) is exist:
                return False
        self._set(key, value, expire)
        return True

    async def set_raw(self, key: Key, value: Value, **kwargs: Any):
        self.store[key] = value

    async def get(self, key: Key, default: Optional[Value] = None) -> Value:
        return await self._get(key, default=default)

    async def get_raw(self, key: Key) -> Value:
        return self.store.get(key)

    async def get_many(self, *keys: Key, default: Optional[Value] = None) -> Tuple[Optional[Value], ...]:
        values = []
        for key in keys:
            val = await self._get(key, default=default)
            if isinstance(val, Bitarray):
                continue
            values.append(val)
        return values

    async def set_many(self, pairs: Mapping[Key, Value], expire: Optional[float] = None):
        for key, value in pairs.items():
            self._set(key, value, expire)

    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[Key]:  # type: ignore
        pattern = pattern.replace("*", ".*")
        regexp = re.compile(pattern)
        for key in dict(self.store):
            if regexp.fullmatch(key):
                yield key

    async def incr(self, key: Key, value: int = 1, expire: Optional[float] = None) -> int:
        value += int(await self._get(key, 0))  # type: ignore
        _expire = None if value != 1 else expire
        self._set(key=key, value=value, expire=_expire)
        return value

    async def exists(self, key: Key) -> bool:
        return await self._key_exist(key)

    async def delete(self, key: Key):
        return await self._delete(key)

    async def _delete(self, key: Key) -> bool:
        if key in self.store:
            del self.store[key]
            await self._call_on_remove_callbacks(key)
            return True
        return False

    async def delete_many(self, *keys: Key):
        for key in keys:
            await self._delete(key)

    async def delete_match(self, pattern: Key):
        async for key in self.scan(pattern):
            await self._delete(key)

    async def get_match(
        self,
        pattern: str,
        batch_size: int = None,
    ) -> AsyncIterator[Tuple[Key, Value]]:  # type: ignore
        async for key in self.scan(pattern):
            value = await self.get(key)
            if not isinstance(value, Bitarray):
                yield key, value

    async def expire(self, key: Key, timeout: float):
        if not await self._key_exist(key):
            return
        value = await self._get(key, default=_missed)
        if value is _missed:
            return
        self._set(key, value, timeout)

    async def get_expire(self, key: Key) -> int:
        if key not in self.store:
            return NOT_EXIST
        expire_at, _ = self.store[key]
        if expire_at is not None:
            return round(expire_at - time.time())
        return UNLIMITED

    async def ping(self, message: Optional[bytes] = None) -> bytes:
        return b"PONG" if message in (None, b"PING") else message  # type: ignore[return-value]

    async def get_bits(self, key: Key, *indexes: int, size: int = 1) -> Tuple[int, ...]:
        array: Bitarray = await self._get(key, default=Bitarray("0"))  # type: ignore
        return tuple(array.get(index, size) for index in indexes)

    async def incr_bits(self, key: Key, *indexes: int, size: int = 1, by: int = 1) -> Tuple[int, ...]:
        array: Optional[Bitarray] = await self._get(key)
        if array is None:
            array = Bitarray("0")
            self._set(key, array)
        result = []
        for index in indexes:
            array.incr(index, size, by)
            result.append(array.get(index, size))
        return tuple(result)

    def _set(self, key: Key, value: Value, expire: Optional[float] = None):
        expire = time.time() + expire if expire else None
        if expire is None and key in self.store:
            expire, _ = self.store[key]
        self.store[key] = (expire, value)
        self.store.move_to_end(key)
        if len(self.store) > self.size:
            self.store.popitem(last=False)

    async def _get(self, key: Key, default: Optional[Value] = None) -> Optional[Value]:
        if key not in self.store:
            return default
        self.store.move_to_end(key)
        expire_at, value = self.store[key]
        if expire_at and expire_at < time.time():
            await self._delete(key)
            return default
        return value

    async def _key_exist(self, key: Key) -> bool:
        return (await self._get(key, default=_missed)) is not _missed

    async def is_locked(self, key: Key, wait: Optional[float] = None, step: float = 0.1) -> bool:
        if wait is None:
            return await self._key_exist(key)
        while wait > 0:
            if not await self._key_exist(key):
                return False
            wait -= step
            await asyncio.sleep(step)
        return await self._key_exist(key)

    async def unlock(self, key: Key, value: Value) -> bool:
        return await self._delete(key)

    async def get_size(self, key: Key) -> int:
        if key in self.store:
            return get_obj_size(self.store[key])
        return 0

    async def slice_incr(self, key: Key, start: int, end: int, maxvalue: int, expire: Optional[float] = None) -> int:
        val_list = await self._get(key)
        count = 0
        new_val = []
        if val_list:
            for val in val_list:
                if start <= val < end:
                    count += 1
                    new_val.append(val)
        if count < maxvalue:
            count += 1
            new_val.append(end)
        self._set(key, new_val, expire=expire)
        return count

    async def set_add(self, key: Key, *values: str, expire: Optional[float] = None):
        val = await self._get(key, default=set())
        val.update(values)
        self._set(key, val, expire=expire)

    async def set_remove(self, key: Key, *values: str):
        val = await self._get(key, default=set())
        val.difference_update(values)
        self._set(key, val)

    async def set_pop(self, key: Key, count: int = 100) -> Iterable[str]:
        values = await self._get(key, default=set())
        _values = []
        for _ in range(count):
            if not values:
                break
            _values.append(values.pop())

        self._set(key, values)
        return _values

    async def close(self):
        self.__remove_expired_stop.set()
        if self.__remove_expired_task:
            await self.__remove_expired_task
            self.__remove_expired_task = None
        self.__is_init = False


class Memory(SerializerMixin, _Memory):
    pass



import time
from contextvars import ContextVar
from functools import wraps
from typing import Any, Dict, Tuple

from cashews._typing import AsyncCallable_T, CallableCacheCondition, Decorator

_spent = ContextVar("spent", default=0.0)


def create_time_condition(limit: float) -> Tuple[CallableCacheCondition, Decorator]:
    def decorator(func: AsyncCallable_T) -> AsyncCallable_T:
        @wraps(func)
        async def _wrapper(*args, **kwargs):
            start = time.perf_counter()
            try:
                return await func(*args, **kwargs)
            finally:
                _spent.set(time.perf_counter() - start)

        return _wrapper

    def condition(result: Any, args: Tuple, kwargs: Dict[str, Any], key: str = "") -> bool:
        return _spent.get() > limit

    return condition, decorator
